"""Executed when sniplet is called as a script

    This file is part of Sniplet.

    Sniplet is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Sniplet is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sniplet.  If not, see <https://www.gnu.org/licenses/>.
    """
from .sniplet import main
main()


"""
    This file is part of Sniplet.

    Sniplet is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Sniplet is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sniplet.  If not, see <https://www.gnu.org/licenses/>.
"""

import click
import os


def read_config_file(filename, verbose):
    """
    Reads the given config file and produces an error if format is wrong

    Args:
        filename(str): File name of file to be opened
        verbose(bool): Control verbose output

    Returns:
        mappings (dict): Holds the mapping of Text Replacements to their Shortcuts
    """

    try:
        f = open(filename, "r")
        lines = f.readlines()
        mappings = {}
        i = 0
        while i < len(lines):
            maplist = []
            if lines[i] == "Short:\n":
                i += 1
                short = lines[i].rstrip("\n")
                i += 1
            else:  # No short tag
                click.echo("Wrong file format!, Aborting....")
                exit(1)
            if lines[i] == "Repl:\n":
                i += 1
                while i < len(lines) and lines[i] != "Short:\n":
                    maplist.append(lines[i])
                    i += 1
                i -= 1
            else:  # No repl tag
                click.echo("Wrong file format!, Aborting....")
                exit(1)

            mappings[short] = maplist  # Add the corresponding lines to the mapping
            i += 1

        if verbose:
            click.echo("Found {} entries in file\n".format(len(mappings)))
            for j in mappings.keys():
                click.echo("Short: {}\n".format(j))
                click.echo("Repl: \n")
                for k in mappings[j]:
                    click.echo(k)
        return mappings
    except FileNotFoundError:   # Exception raised if file not found
        click.echo("The specified file was not found!, Aborting....")
        exit(1)


def create_config_file():
    """
    Creates a config file from given user input and produces an error if format is wrong
    """

    click.echo("Enter the file name: ")
    filename = input()
    mappings = {}

    while True:
        click.echo("\n1.Add new replacement\n2.Delete replacement\n3.Show all replacements\n4.Exit\n")
        choice = input()
        if choice == "1":
            click.echo("Enter the shortcut: ")
            short = input()
            click.echo("Enter the Replacement: ")
            repl = input()
            mappings[short] = repl

        elif choice == "2":
            click.echo("Enter the shortcut for the replacement to be deleted: ")
            short = input()
            try:
                del mappings[short]
            except KeyError:
                click.echo("Record not found\n")

        elif choice == "3":
            for i in mappings.keys():
                click.echo("\nShortcut: {} -> Replacement: {}".format(i, mappings[i]))

        elif choice == "4":
            break

        else:
            click.echo("Please choose the correct option!\n")

    click.echo("Writing to file {} .....".format(filename))

    try:
        os.mkdir("./config-files/")
    except FileExistsError:
        pass

    f = open("./config-files/"+filename, "w")
    for i in mappings.keys():
        f.write("Short:\n")
        f.write(i + "\n")
        f.write("Repl:\n")
        f.write(mappings[i] + "\n")
    click.echo("File created successfully!")



"""
    This file is part of Sniplet.

    Sniplet is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Sniplet is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sniplet.  If not, see <https://www.gnu.org/licenses/>.
"""

from ._helper import *
from pynput import keyboard
from pynput.keyboard import Key, Controller
import click


class Snipper:
    """
    Class which holds the data and functions required for snipper

        Attributes:
            current_input (list): Stores the input till space is pressed
            lis (bool): Flag to control listening
            running (bool): Flag to control running of the program
            kb (Keyboard.controller): Controller used to create keyboard actions
            listener (Keyboard.listener): Listens to the keyboard
            verbose (bool): Printing verbose outputs
            mappings (dict): Holds the mapping of Text Replacements to their Shortcuts
    """

    def __init__(self, filename, verbose):
        self.current_input = []
        self.lis = False
        self.running = False
        self.kb = Controller()
        self.listener = 0
        self.verbose = verbose
        self.mappings = read_config_file(filename, verbose)

    def on_press(self, key):
        """
        Text replacement is done. Called on press of a key
            Args:
                key (Keys.key): Key being pressed
            Returns:
                bool: Stop or continue listening
        """
        keys = set(self.mappings.keys())
        try:
            self.current_input.append(key.char)
        except AttributeError:
            self.listener.stop()
            self.running = False
            if key == Key.space:
                curstr = ''.join([str(elem) for elem in self.current_input])
                if self.verbose:
                    print(curstr)
                if curstr in keys:
                    if self.verbose:
                        click.echo("Replacing {} with {}".format(curstr, self.mappings[curstr]))
                    for i in range(len(curstr) + 1):
                        self.kb.press(Key.backspace)
                        self.kb.release(key.backspace)
                    for i in range(len(self.mappings[curstr])):
                        self.kb.type(self.mappings[curstr][i])
                    self.kb.press(Key.backspace)
                    self.kb.release(key.backspace)
                    self.kb.press(Key.space)
                    self.kb.release(key.space)
                    self.current_input = []
                self.current_input = []
            elif key == Key.backspace and self.current_input:
                self.current_input.pop()

    def on_release(self, key):
        """
            Called on release of a key (Not working properly)
                Args:
                    key (Keys.key): Key being pressed
                Returns:
                    bool: Stop or continue listening
        """
        if key == Key.esc:
            self.lis = False
            self.running = False
            self.listener.stop()
            return False

    def run(self):
        """Starts the listening process"""
        self.lis = True
        self.running = False
        self.listener = keyboard.Listener(on_press=lambda event: self.on_press(event),
                                          on_release=lambda event: self.on_release(event))  # Creating a listener
        while self.lis: # Main Loop
            if not self.running:
                self.listener.stop()
                self.listener = keyboard.Listener(on_press=lambda event: self.on_press(event),
                                                  on_release=lambda event: self.on_release(event)) # Start new listener
                self.running = True
                self.listener.start() # Starting the thread
                self.listener.join() # Joining the threads in a blocking way



"""
    This file is part of Sniplet.

    Sniplet is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Sniplet is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sniplet.  If not, see <https://www.gnu.org/licenses/>.
    """
import click
from ._snipper import *

__version__ = "1.0.7"


@click.command()
@click.option('--verbose', is_flag=True, help="Print Verbose Messages")
@click.option('--config', is_flag=True, help="Configure the Snippets")
@click.option('--version', is_flag=True, help="Get the version number")
@click.option('--usrfile', required="false", default="", help="Start the snipper with user defined config file as "
                                                              "filename ( "
                                                              "Follow file "
                                                              "convention in --help)")
def main(verbose, config, version, usrfile):
    """
    Sniplet

    A lightweight, threaded, multiplatform text replacement tool which works on Windows, MacOS and Linux

    Note: Requires keyboard access on MacOS and X Server on Linux (for pynput)

    To configure replacement file within the terminal use --config. To use your own config file please mention it
    with --start (Path Required) and follow the given specification

    File Specification:

    Short:

    [Shortcut]

    Repl:

    [Replacement]

    A Sample Specification is provided with sample_config.txt inside config-files

    !!Turn off the listener from the terminal using the keyboard interrupt to avoid unnecessary leakage of text!!

    Author: Veeraraghavan Narasimhan

    Email: virunarasimhan25@gmail.com

    Licensed under GNU General Public License 3
    """

    click.echo("Licensed under GNU General Public License 3\n")

    if usrfile != "":
        snip = Snipper(usrfile, verbose)
        click.echo("Started Listening!\n")
        snip.run()

    elif config:
        create_config_file()

    elif version:
        click.echo(__version__)

    elif usrfile == "":
        click.echo("Enter the filename you created using config: ")
        filename = input()
        filename = "./config-files/" + filename
        snip = Snipper(filename, verbose)
        click.echo("Started Listening!\n")
        snip.run()



"""Setup.py: Setup tools control

    This file is part of Sniplet.

    Sniplet is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Sniplet is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sniplet.  If not, see <https://www.gnu.org/licenses/>.
"""

import re
from setuptools import setup

version = re.search('^__version__\s*=\s*"(.*)"', open('sniplet/sniplet.py').read(), re.M).group(1)
setup(
    name="sniplet",
    packages=["sniplet"],
    entry_points={
        "console_scripts": ['sniplet = sniplet.sniplet:main']
    },
    version=version,
    description="Python command line application for text replacement",
    author="Veeraraghavan-Narasimhan",
    author_email="virunarasimhan25@gmail.com",
    url="https://github.com/veera2508",
    install_requires=["pynput", "click"]
)
